/**
 * Copyright 2014 Nikita Koksharov, Nickolay Borbit
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.redisson;

import org.redisson.connection.LoadBalancer;
import org.redisson.connection.RoundRobinLoadBalancer;

public class BaseMasterSlaveServersConfig<T extends BaseMasterSlaveServersConfig<T>> extends BaseConfig<T> {

    /**
     * Сonnection load balancer for multiple Redis slave servers
     */
    private LoadBalancer loadBalancer = new RoundRobinLoadBalancer();

    /**
     * Redis 'slave' servers subscription (pub/sub) connection pool size for <b>each</b> slave node
     */
    private int slaveSubscriptionConnectionPoolSize = 25;

    /**
     * Redis 'slave' servers connection pool size for <b>each</b> slave node
     */
    private int slaveConnectionPoolSize = 100;

    /**
     * Redis 'master' server connection pool size
     */
    private int masterConnectionPoolSize = 100;

    /**
     * Redis 'slave' server reconnection attempt timeout
     * used then server excluded from the list of available slave nodes
     * due to reach limit of sequential unsuccessful execution attempts
     *
     * @see #slaveFailedAttempts
     *
     */
    private int slaveReconnectionTimeout = 3000;

    /**
     * Redis 'slave' server will be excluded from the list of available slave nodes
     * when sequential unsuccessful execution attempts of any Redis command on slave node reaches <code>slaveFailedAttempts</code>
     */
    private int slaveFailedAttempts = 3;

    public BaseMasterSlaveServersConfig() {
    }

    BaseMasterSlaveServersConfig(T config) {
        super(config);
        setLoadBalancer(config.getLoadBalancer());
        setMasterConnectionPoolSize(config.getMasterConnectionPoolSize());
        setSlaveConnectionPoolSize(config.getSlaveConnectionPoolSize());
        setSlaveSubscriptionConnectionPoolSize(config.getSlaveSubscriptionConnectionPoolSize());
        setSlaveFailedAttempts(config.getSlaveFailedAttempts());
        setSlaveReconnectionTimeout(config.getSlaveReconnectionTimeout());
    }

    /**
     * Redis 'slave' servers connection pool size for <b>each</b> slave node
     * Default is 100
     *
     * @param slaveConnectionPoolSize
     * @return
     */
    public T setSlaveConnectionPoolSize(int slaveConnectionPoolSize) {
        this.slaveConnectionPoolSize = slaveConnectionPoolSize;
        return (T)this;
    }
    public int getSlaveConnectionPoolSize() {
        return slaveConnectionPoolSize;
    }

    /**
     * Redis 'master' server connection pool size
     * Default is 100
     *
     * @param masterConnectionPoolSize
     * @return
     */
    public T setMasterConnectionPoolSize(int masterConnectionPoolSize) {
        this.masterConnectionPoolSize = masterConnectionPoolSize;
        return (T)this;
    }
    public int getMasterConnectionPoolSize() {
        return masterConnectionPoolSize;
    }

    /**
     * Сonnection load balancer to multiple Redis slave servers.
     * Uses Round-robin algorithm by default
     *
     * @param loadBalancer
     * @return
     *
     * @see org.redisson.connection.RoundRobinLoadBalancer
     * @see org.redisson.connection.BaseLoadBalancer
     */
    public T setLoadBalancer(LoadBalancer loadBalancer) {
        this.loadBalancer = loadBalancer;
        return (T)this;
    }
    public LoadBalancer getLoadBalancer() {
        return loadBalancer;
    }

    /**
     * Redis 'slave' servers subscription connection pool size for <b>each</b> slave node
     * Default is 25
     *
     * @param slaveSubscriptionConnectionPoolSize
     * @return
     */
    public T setSlaveSubscriptionConnectionPoolSize(int slaveSubscriptionConnectionPoolSize) {
        this.slaveSubscriptionConnectionPoolSize = slaveSubscriptionConnectionPoolSize;
        return (T)this;
    }
    public int getSlaveSubscriptionConnectionPoolSize() {
        return slaveSubscriptionConnectionPoolSize;
    }

    /**
     * Redis 'slave' server reconnection attempt timeout
     * used then server excluded from the list of available slave nodes
     * due to reach limit of sequential unsuccessful execution attempts
     *
     * @see #slaveFailedAttempts
     *
     */
    public T setSlaveReconnectionTimeout(int slaveRetryTimeout) {
        this.slaveReconnectionTimeout = slaveRetryTimeout;
        return (T)this;
    }
    public int getSlaveReconnectionTimeout() {
        return slaveReconnectionTimeout;
    }

    /**
     * Redis 'slave' server will be excluded from the list of available slave nodes
     * when sequential unsuccessful execution attempts of any Redis command on slave node reaches <code>slaveFailedAttempts</code>
     */
    public T setSlaveFailedAttempts(int slaveFailedAttempts) {
        this.slaveFailedAttempts = slaveFailedAttempts;
        return (T)this;
    }
    public int getSlaveFailedAttempts() {
        return slaveFailedAttempts;
    }

}
